import { UnpackError } from "./errors";

export const Heartbeat = Symbol("Heartbeat");

export interface DataMessage {
  _: unknown;
}

export function pack(data: unknown): string {
  return JSON.stringify({ _: data }) as string;
}

/**
 * Returns the data in the message or the Heartbeat symbol
 *
 * @param data - The JSON string to be deserialized.
 * @returns The deserialized data or the Heartbeat symbol
 * @throws {ManagedSocketError} If the JSON string cannot be parsed or if the message shape is invalid.
 */
export function unpack(data: string): { data: unknown } | typeof Heartbeat {
  try {
    const parsedData = JSON.parse(data);

    // If the parsed data is a number, it's a heartbeat
    if (typeof parsedData === "number") {
      return Heartbeat;
    }

    // Otherwise, the shape should be an object with a _ property holding the data
    if (
      typeof parsedData === "object" &&
      parsedData !== null &&
      "_" in parsedData &&
      typeof parsedData._ !== "undefined"
    ) {
      return { data: parsedData._ };
    }

    throw new UnpackError("Unpack Error: Invalid message shape");
  } catch (error) {
    if (error instanceof Error) {
      throw new UnpackError("Unpack Error: Failed to deserialize", error);
    }

    // Raise unknown exceptions
    throw error;
  }
}
