import { SocketError } from "./errors";
import { pack, unpack } from "./messages";

const THIRTY_SECONDS = 30 * 1000;

type SocketDataHandler = (data: unknown) => void | Promise<void>;
type SocketErrorHandler = (error: SocketError) => void;
type SocketConnectedHandler = () => void;
type SocketDisconnectedHandler = ({
  wasClean,
  code,
  reason,
}: {
  wasClean: boolean;
  code: number;
  reason: string;
}) => void;

interface HandlerDefinitions {
  onMessage?: SocketDataHandler;
  onError?: SocketErrorHandler;
  onConnected?: SocketConnectedHandler;
  onDisconnected?: SocketDisconnectedHandler;
}

export class ManagedSocket {
  public readonly url: string;
  public socket: WebSocket;

  public onMessage: SocketDataHandler = () => {};
  public onError: SocketErrorHandler = () => {};
  public onConnected: SocketConnectedHandler = () => {};
  public onDisconnected: SocketDisconnectedHandler = () => {};

  private _reconnectAttempts = 0;
  private _messageQueue: unknown[] = [];
  private _debugMode = false;
  private _offline = false;

  constructor(url: string, handlers: HandlerDefinitions = {}) {
    this.url = url;

    this.onMessage = handlers.onMessage ?? this.onMessage;
    this.onError = handlers.onError ?? this.onError;
    this.onConnected = handlers.onConnected ?? this.onConnected;
    this.onDisconnected = handlers.onDisconnected ?? this.onDisconnected;

    this.socket = this.connect();

    setInterval(() => {
      this.sendHeartbeat();
    }, THIRTY_SECONDS);

    window.addEventListener("offline", () => {
      this._offline = true;

      if (this.isConnected) {
        this.debugMessage("Internet connection lost, closing socket...");
        this.close();
      }
    });

    window.addEventListener("online", () => {
      this._offline = false;

      if (!this.isConnected) {
        this.debugMessage(
          "Internet connection restored, reconnecting socket...",
        );
        this.attemptReconnection();
      }
    });
  }

  private sendHeartbeat() {
    if (this.socket?.readyState === WebSocket.OPEN) {
      this.socket.send(JSON.stringify(Date.now()));
    }
  }

  private debugMessage(...args: unknown[]) {
    if (this._debugMode) {
      const output = ["[Socket]", ...args];
      console.log(...output);
    }
  }

  private sendQueuedMessages() {
    // Make a copy of the queue  and clear it first.
    // this way if the messages fail to send again,
    // they just get requeued
    const messagesToSend = [...this._messageQueue];
    this._messageQueue = [];
    for (const message of messagesToSend) {
      this.send(message);
    }
  }

  private attemptReconnection() {
    if (this._offline) {
      this.debugMessage("Browser offline; skipping reconnection attempt");
      return;
    }

    const delay = Math.min(2000 ** this._reconnectAttempts, THIRTY_SECONDS);
    this.debugMessage(`Attempting to reconnect in ${delay}ms`);
    setTimeout(() => {
      this._reconnectAttempts++;
      this.connect();
    }, delay);
  }

  private connect() {
    const newSocket = new WebSocket(this.url);

    // When the connection is established, run the handler
    // and deliver any queued messages
    newSocket.addEventListener("open", () => {
      this._reconnectAttempts = 0;
      this.onConnected();
      this.sendQueuedMessages();
    });

    // When the connection closes, run the handler and possibly
    // attempt a reconnection if there is still an internet connection
    newSocket.addEventListener("close", ({ wasClean, code, reason }) => {
      this.debugMessage("WebSocket closed", {
        wasClean,
        code,
        reason,
      });

      this.onDisconnected({
        wasClean,
        code,
        reason,
      });

      if (!this._offline) {
        this.debugMessage("Attempting automatic reconnection");
        return this.attemptReconnection();
      }
    });

    newSocket.addEventListener("error", (error) => {
      this.debugMessage("WebSocket error", error);

      const socketError = new SocketError(
        "Socket error",
        new Error(error.currentTarget?.toString()),
      );

      this.onError(socketError);
    });

    newSocket.addEventListener("message", (messageEvent) => {
      const data = unpack(messageEvent.data);

      if (data === null) {
        return;
      }

      this.debugMessage("Received message", data);
      this.onMessage(data);
    });

    this.socket = newSocket;
    return newSocket;
  }

  get isConnected() {
    return this.socket?.readyState === WebSocket.OPEN;
  }

  debug(debugModeEnabled: boolean) {
    this._debugMode = debugModeEnabled;
  }

  close() {
    if (this.socket?.readyState !== WebSocket.OPEN) {
      throw new SocketError("Cannot close a socket that is not open");
    }

    this.socket.close();
  }

  send(data: unknown) {
    if (!this.socket || this.socket.readyState !== WebSocket.OPEN) {
      this.debugMessage("WebSocket is not open. Queuing message.");
      this._messageQueue.push(data);
      return;
    }

    this.socket.send(pack(data));
  }
}
